import collections, importlib, math, os, pkgutil, sys
import numpy as np
from support_modules import DTError, DTUtils
import loaders
#https://youtu.be/wQHZL9ePMKk
from loaders import *
        

class DTData:
    
    
    def __init__(self, samples=None, loader=None, loader_file = None, files=None, sa=None, fa=None, auto_load=True):

        self.mask = None

        if loader==None:
            DTError.err("DTData: A loader function must be specified, but was not.")

        if files==None:
            DTError.err("DTData: At least one file must be specified, but none were.")

        #this should be coming in as a list, but just in case
        if type(files) != list:
            files = DTUtils.listify( files )
        self.files = files
        self.loader_file = loader_file
        self.loader = self.dirty_loader_things(loader)
        self.samples = samples
        self.sa = sa
        self.fa = fa
        
        if auto_load:
            self.load_data()
    
    
    def load_data(self):
        [self.samples, self.sa, self.fa] = self.loader(self.files)
    
    
    def apply_mask(self, index_array):
        if self.mask is None:
            return index_array
        else:
            return np.logical_and(self.mask, index_array)
    
    
    def slice_train_val_test(self, train_proportion, val_proportion, test_proportion, classify_over, return_test_indices=False ):
        if (train_proportion + val_proportion + test_proportion) == 100: #must have specified percent, not proportions
            train_proportion /= 100.0
            val_proportion /= 100.0
            test_proportion /= 100.0
        
        if np.abs(train_proportion + val_proportion + test_proportion - 1.0) > .0001:
            DTError.err("DTData: Train proportion + validation proportion + test proportion needs to equal 1.0!")
        
        class_labels = self.sa[classify_over] # TK: insert appropriate error checking here
         # including changing categorical labels to 0, 1, 2, etc... format in here
        
        training_samples = np.empty([0], dtype = int)
        validation_samples = np.empty([0], dtype = int)
        testing_samples = np.empty([0], dtype = int)
        
        n_training_samples = []
        n_validation_samples = []
        n_testing_samples = []
        # TK: make this optional if you DON'T want to equate all classes
        for current_label in np.unique(class_labels):
            # TK: make less gross
            current_samples = np.asarray(np.where(self.apply_mask(class_labels == current_label))).flatten()
            n_training_samples.append( math.floor(current_samples.shape[0] * train_proportion) )
            n_validation_samples.append( math.floor(current_samples.shape[0] * val_proportion) )
            n_testing_samples.append( math.floor(current_samples.shape[0] * test_proportion) )

        n_training_samples = int(min( n_training_samples ))
        n_validation_samples = int(min( n_validation_samples ))
        n_testing_samples = int(min( n_testing_samples ))
        
        for current_label in np.unique(class_labels):
            # TK: use np.setdiff1d rather than casting back and forth
            # TK: change variable names to be more representative
            current_samples = np.asarray(np.where(self.apply_mask(class_labels == current_label))).flatten()
            current_training_samples = np.random.choice(current_samples, n_training_samples, replace = False)
            current_samples = list( set(current_samples) - set(current_training_samples) )
            current_validation_samples = np.random.choice(current_samples, n_validation_samples, replace = False)
            current_samples = list( set(current_samples) - set(current_validation_samples) )
            if test_proportion > .000001:
                current_testing_samples = np.random.choice(current_samples, n_testing_samples, replace = False)
            else:
                current_testing_samples = np.empty([0], dtype = int)

            training_samples = np.concatenate((training_samples, current_training_samples))
            validation_samples = np.concatenate((validation_samples, current_validation_samples))
            testing_samples = np.concatenate((testing_samples, current_testing_samples))

        training_data = self.samples[training_samples,:]
        validation_data = self.samples[validation_samples, :]
        test_data = self.samples[testing_samples,:]
        
        training_labels = class_labels[training_samples]
        validation_labels = class_labels[validation_samples]
        test_labels = class_labels[testing_samples]
        
        if return_test_indices:
            return training_data, validation_data, test_data, training_labels, validation_labels, test_labels, testing_samples
        else:
            return training_data, validation_data, test_data, training_labels, validation_labels, test_labels
    
    
    def mask_sa(self, sa_name, keep_values):
        # keep_values may be iterable or not
        if isinstance(keep_values, collections.Iterable):
            keep_indices = np.asarray([item in keep_values for item in self.sa[sa_name]])
        else: 
            keep_indices = np.asarray(self.sa[sa_name] == keep_values)
        self.mask = keep_indices
        
        
    def dirty_loader_things(self, loader):
        if self.loader_file:
            if os.path.isfile(self.loader_file):
                loader_dir, loader_file_name = os.path.split(self.loader_file)
                sys.path.append(loader_dir)
                loader_import_name = os.path.splitext(loader_file_name)[0]
                loader_module = importlib.import_module(loader_import_name)
                if getattr(loader_module,loader,None):
                    loader = getattr(loader_module,loader)
                else: 
                    DTError.err('The loader file named {} does not appear to contain a function named {}.'.format(self.loader_file,loader))
            else: 
                DTError.err('Unable to locate a loader file named {}. Double check paths, etc.'.format(self.loader_file))
        else:
            for loader_file_name in pkgutil.iter_modules(loaders.__path__):
                if getattr(eval('loaders.'+loader_file_name[1]),str(loader),None):
                    loader = getattr(eval('loaders.'+loader_file_name[1]),loader)
                
        return loader
    
    
    @classmethod
    def validate_arguments( cls, args ):
        # this is the laziest and dumbest possible version of a method like this, which really should inspect the arguments and give some intelligent
        # warning messages and so forth. But it'll do for now to get us off the ground
        try:
            test_obj = cls( **args )
        except:
            global dt_debug_mode
            if 'dt_debug_mode' not in globals():
                dt_debug_mode = True # TK: change to false when we go into public production
            
            if dt_debug_mode:
                DTError.err( 'DTData: validate_arguments() failed' )
            else:
                return False
        
        return True
    
    
    @classmethod
    def pymvpa_wrap( cls, in_data, in_labels):
        from mvpa2.datasets.base import Dataset
        #KK: I don't love this, but the MRJ mandate says we go as far as possible before bringing in PyMVPA functions
        
        out_data = Dataset(in_data)
        out_data.sa['targets'] = in_labels
        return out_data
    
    
    # this next method (to_categorical) is copied straight from Keras, but re-implementing it so we don't have to have it as a dependency
    # n.b. Keras is MIT License so here is the copyright/license information from it, to satisfy those terms:
    # 
    # COPYRIGHT
    # 
    # All contributions by Francois Chollet:
    # Copyright (c) 2015 - 2018, Francois Chollet.
    # All rights reserved.
    # 
    # All contributions by Google:
    # Copyright (c) 2015 - 2018, Google, Inc.
    # All rights reserved.
    # 
    # All contributions by Microsoft:
    # Copyright (c) 2017 - 2018, Microsoft, Inc.
    # All rights reserved.
    # 
    # All other contributions:
    # Copyright (c) 2015 - 2018, the respective contributors.
    # All rights reserved.
    # 
    # Each contributor holds copyright over their respective contributions.
    # The project versioning (Git) records all such contribution source information.
    # 
    # LICENSE
    # 
    # The MIT License (MIT)
    # 
    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:
    # 
    # The above copyright notice and this permission notice shall be included in all
    # copies or substantial portions of the Software.
    # 
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    # SOFTWARE.
    # 
    @classmethod
    def to_categorical( cls, y, num_classes=None):
        """Converts a class vector (integers) to binary class matrix.
        E.g. for use with categorical_crossentropy.
        # Arguments
            y: class vector to be converted into a matrix
                (integers from 0 to num_classes).
            num_classes: total number of classes.
        # Returns
            A binary matrix representation of the input.
        """
        y = np.array(y, dtype='int').ravel()
        if not num_classes:
            num_classes = np.max(y) + 1
        n = y.shape[0]
        categorical = np.zeros((n, num_classes))
        categorical[np.arange(n), y] = 1
        return categorical
    
    
    @classmethod
    def train_val_test_to_categorical( cls, train, val=None, test=None ):
        n_classes_train = (np.unique(train)).shape[0]
        train = DTData.to_categorical( train, n_classes_train )
        
        if not (val is None):
            n_classes_val = (np.unique(val)).shape[0]
            if n_classes_val != n_classes_train:
                my_message = [  "DTData: Number of training classes doesn't match number of validation classes ",
                                "        in DTData.train_val_test_to_categorical()" ]
                DTError.err( my_message )
            val = DTData.to_categorical( val, n_classes_val )
        
        if not (test is None):
            test = DTData.to_categorical( test, n_classes_train ) #test doesn't necessarily have to have all the classes
        
        return (train, val, test)
