# DTJob (class) #

---

The primary class that maintains information about a job (or list of jobs) to run. The most common usage of this class would probably be indirectly, e.g., by passing a JSON-based job configuration file to `delineate.py`, which basically just creates `DTJob` objects out of each file it is given and runs them in sequence. So if you are using `delineate.py` and specifying your jobs in JSON config files, you don't need to know too much about the inner workings of this class. Just see the documentation for the JSON-format job config files and `delineate.py`, and you should be all set.

However, if you are importing these modules manually and writing your own Python code, you may find yourself creating `DTJob` objects. See `dt_design_looper_example.py` for a basic script that does this.

The main thing to know in either case is that `DTJob` essentially maintains a structure of parameters (or hyperparameters, in deep learning parlance) that define a given analysis using a given dataset. Eventually, when the `run()` method is called, it will use those (hyper-)parameters to create `DTAnalysis`, `DTData`, and `DTModel` objects. So you can think of `DTJob` as kind of an umbrella container for those objects, although in the current implementation it doesn't hang on to them for any length of time -- those objects are just created long enough to run the analysis, and then destroyed. Another way to think of `DTJob` objects is as the thawed, usable form of the information frozen in JSON-based job config files.

## Attributes ##


**job_file** *(string)*: A filename or path to a JSON-format job configuration file that will be used to create the rest of the `DTJob` object. This is the only attribute that can be specified when creating a new `DTJob` object via the `__init__()` method; if that's the way you are going to create your `DTJob` object (which is essentially what `delineate.py` does), then you won't have to fiddle with any of the other attributes. In this case, all you'd have to do is something like:

> `import DTJob`

> `my_job = DTJob.DTJob( 'my_job_file.json' )`

> `my_job.run()`

Although if that's **all** you want to do, you could probably just run `delineate.py` and pass it `my_job_file.json` in the first place.

If you are manually creating your `DTJob` object, then this attribute is not mandatory to specify; in `__init()__`, it defaults to `None`. So you can also just do:

> `import DTJob`

> `my_job = DTJob.DTJob()`

... and then fill in the rest of the attributes later, before you run the job.

Another option would be to load a basic job configuration in via a JSON file, but then tweak the job structure manually before you actually run. This is fundamentally not too different from specifying the whole job structure yourself in code, but it might save you a few lines of code if you typically use a very similar configuration (which can be saved in the JSON file and loaded in to create a template `DTJob` object) and just want to tweak a few things in your Python code.

Note that **job_file** is just going to be passed along to Python's regular `open()` function eventually, so it can be a full path or a bare filename or whatever, just as long as it is accessible from wherever you're running.

<br>

**job_structure** (*list of dictionaries*): This is the actual data structure representing what is going to get run (and with what data), either specified in **job_file** or in code that you write.

It follows the same basic format as the JSON-format job config files, so for a breakdown of all the stuff that can go into this data structure, see the documentation for those files.

Essentially, though, **job_structure** is a list of Python dictionaries. (However, if you make it a single Python dictionary, `DTJob` should generally be nice and turn it into a single-item list as necessary... but you should probably think of it and use it as a list of dictionaries to avoid problems down the line.) Each dictionary contains the information needed to run one job. The JSON job config files can contain either a single job or a list of jobs, so loading the contents of one job file into one **job_structure** maintains a 1:1 correspondence between actual job files and `DTJob` objects.

Each dictionary contains three main keys: `data`, `analysis`, and `model`. The values corresponding to those keys will in turn be other dictionaries; each of these dictionaries contains the (hyper-)parameters necessary to instantiate a `DTData`, a `DTAnalysis`, and a `DTModel` object, respectively. Again, see the docs for the config files for more details on what should go into those.

As suggested above, see `dt_design_looper_example.py` for some ideas as to how you might go about creating this structure manually in Python code.

Also, as noted above in the description of the **job_file** attribute, if you wanted to initially create a `DTJob` object using an actual JSON job config file, then tweak the **job_structure** attribute manually before running, that would be perfectly allowable. The **job_file** parameter is automatically converted into a **job_structure** if you pass in a job file when creating a brand-new `DTJob` object, so you can assume if you create a `DTJob` object from a file that the **job_structure** will exist when the object is done being initialized. Alternately, if you add in a **job_file** attribute after creating the `DTJob` object for some reason, you will need to call the `reload()` method (see below) to turn that file into a **job_structure**. In any case, whatever you want to do to modify the **job_structure** before calling the `DTJob.run()` method is totally up to you.


<br><br><br><br><br>


notes (to clean up later):

* does not run automatically -- have to explicitly call run() if not using delineate.py or similar
* default is to initialize basically everything to None
* OK to specify job structure directly and call run() (without giving a job file) if you know what you're doing
* for validate_job_structure() -- currently there is an option to pass in an arbitrary job structure. We don't really ever exercise this option -- I think it is still around from when we first tried to implement this as a class method? Anyway, we're leaving it this way for now even if it's kinda vestigial -- might come in handy later if we ever make an explicit pre-run validation feature
* Note the general idea for validation here is to pass off to DTModel, DTAnalysis, DTData and have them validate their own parameters -- currently the implementation is just to try to instantiate each one and report an error if it doesn't work (and throw away the result if it does work), but this could get smarter in the future.



below: copied from python comments, needs formatting/updating

A class to manage multiple runs of analyses, based on an input config text file (JSON format).

    Attributes:
    
    last_loaded_job_file (type: string) -- this gets updated with the value of job_file when it is successfully loaded; used to issue a warning if someone
        loads a job file, then changes the job_file attribute but does not reload
    
    Methods:
    
    reload() -- if 'job_file' changes, call this to recreate 'job_structure'
    run() -- run whatever is currently loaded
    validate_job_structure() -- do a basic check to see if the essential structural elements of the job are present (and nothing is present that DTJob
        doesn't understand). This won't check e.g. to see if parameters passed into the ultimate model are valid, if data files exist, etc.... it's just 
        making sure that the job file is formatted properly.
    validate_job_structure_onedict() -- helper method for validate_job_structure(), should not need to call this directly unless you're doing something 
        weird.
